/* $Id: llist.h 9 2007-10-25 12:37:14Z csg $ */
#ifndef _LLIST_H_
#define _LLIST_H_

typedef struct t_llist t_llist;
typedef struct t_record t_iterator;

/* Returns a pointer to an new created, empty linked list. */
t_llist *llist_create(void);

/* Will free all memory related to the linked list `l'. If `user_data_destroy' is not NULL it will be called once for
   each data element in the linked list. */
void llist_destroy(t_llist *l, void (*user_data_destroy)(void *data));

/* Insert the element `data' first in the list `l'. */
void llist_insert_first(t_llist *l, void *data);

/* Insert the element `data' last in the list `l'. */
void llist_insert_last(t_llist *l, void *data);

/* Insert the element `data' before the location of `where' in the list `l'. If `where' is not found, then `data' will be
   inserted first. */
void llist_insert_at(t_llist *l, void *data, void *where);

/* 'llist_concatenate' concatenates the list l1 and l2. After the call l2 is no longer a valid pointer, and all its former
   elements are contained in l1. The element order is preserved. The elements from l2 are apppended at end of those in l1 */
void llist_concatenate(t_llist *l1, t_llist *l2);

/* Returns a bi-directional iterator to the first element in the list `l'. */
t_iterator *llist_first(t_llist *l);

/* Returns a bi-directional iterator to the last element in the list `l'. */
t_iterator *llist_last(t_llist *l);

/* Returns the next iterator of the list where the iterator `it' belongs. If `it' refers to the last element in ths list
   then it will return NULL. */
t_iterator *llist_next(t_iterator *it);

/* Returns the previous iterator of the list where the iterator `it' belongs. If `it' refers to the first element in ths list
   then it will return NULL. */
t_iterator *llist_prev(t_iterator *it);

/* Removes an element refered by the iterator `it' in the list `l'. The return value is an iterator refering to the next
   element in the list or NULL if this was the last element. */
t_iterator *llist_remove_at(t_llist *l, t_iterator *it);

/* Removes an element that contains the data pointer `data' if it exist, else do nothing. Returns non-0 if `data' existed. */
int llist_remove(t_llist *l, void *data);

/* Returns the data object in the list element refered by the iterator `it'. */
void *llist_data(t_iterator *it);

#endif
